package com.analogbit.puzzletools.client.slidingblocks.state;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.PriorityQueue;

import com.google.gwt.user.client.Timer;

public class BlockPuzzleSolver extends Timer {
	public static final int STEPS_PER_TIMETEST = 50;
	public static final int SLEEP_SLICE = 10;
	public static final int TIME_SLICE = 100;
	public static final int PUZZLE_FAILED = -1;
	public static final int PUZZLE_CANCELLED = -2;

	private ArrayList<BlockPuzzleSolverListener> listeners;
	//private LinkedList<Grid> pending;
	private HashSet<Grid> seen;
	private int currentDepth;
	private GoalState goalState;
	private Grid answer;
	private long positionsEvaluated;
	private long duplicatePositions;
	private PriorityQueue<Grid> pending;
	private ArrayList<Grid> temp;

	public BlockPuzzleSolver(Grid gridToSolve, GoalState goalState) {
		this.goalState = goalState;
		this.listeners = new ArrayList<BlockPuzzleSolverListener>();
		this.pending = new PriorityQueue<Grid>();
		this.seen = new HashSet<Grid>();
		this.currentDepth = 0;
		this.positionsEvaluated = 0;
		this.duplicatePositions = 0;
		this.temp = new ArrayList<Grid>();

		gridToSolve.recomputeBlockAliases(goalState);
		addToPendingIfNotSeen(gridToSolve);
	}

	public void addListener(BlockPuzzleSolverListener listener) {
		listeners.add(listener);
	}

	public boolean removeListener(BlockPuzzleSolverListener listener) {
		return listeners.remove(listener);
	}

	public void beginSolving() {
		this.scheduleRepeating(SLEEP_SLICE);
	}

	public void cancelSolving() {
		this.cancel();
		firePuzzleSolved(PUZZLE_CANCELLED);
	}

	@Override
	public void run() {
		if (answer != null) {
			firePuzzleSolved(currentDepth);
			return;
		}

		long startTime = System.currentTimeMillis();
		do {
			int stepsRemaining = STEPS_PER_TIMETEST;
			while (stepsRemaining-- > 0) {
				Grid g = pending.remove();
				positionsEvaluated++;

				int distFromGoal = goalState.distanceFromGoal(g); 
				if (distFromGoal == 0) {
					answer = g;
					firePuzzleSolved(currentDepth);
					return;
				} 

				addChildrenToPending(g);
				
				if (pending.isEmpty()) {
					firePuzzleSolved(PUZZLE_FAILED);
					return;
				}
			}
		} while ((System.currentTimeMillis() - startTime) < TIME_SLICE);
		
		currentDepth = pending.peek().getDepth();
		fireInformationUpdate();
	}

	private void addChildrenToPending(Grid g) {
		g.getAllValidPermutations(temp);
		int numChildren = temp.size();
		for (int i = 0; i<numChildren; i++) {
			Grid newGrid = temp.get(i);
			if (newGrid != null) {
				addToPendingIfNotSeen(newGrid);
			}
		}
		temp.clear();
	}

	private void addToPendingIfNotSeen(Grid newGrid) {
		if (seen.add(newGrid)) {
			newGrid.setDistanceFromGoal(goalState.distanceFromGoal(newGrid));
			pending.add(newGrid);
		} else {
			duplicatePositions++;
		}
	}

	private void firePuzzleSolved(int depth) {
		this.cancel();
		
		//cleanup memory immediately
		if (pending != null) {
			pending.clear();
			pending = null;
		}
		if (seen != null) {
			seen.clear();
			seen = null;
		}
		
		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).puzzleComplete(depth);
		}
	}
	
	private void fireInformationUpdate() {
		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).puzzleStatusUpdate(currentDepth, pending.size(), positionsEvaluated, duplicatePositions);
		}
	}
	
	public ArrayList<SolutionStep> getAnswer() {
		if (answer == null) {
			return null;
		}
		ArrayList<SolutionStep> result = new ArrayList<SolutionStep>();
		getAnswerRecursive(result, answer);
		return result;
	}

	private static void getAnswerRecursive(ArrayList<SolutionStep> list, Grid g) {
		if (g == null) {
			return;
		}
		
		getAnswerRecursive(list, g.getParentGrid());
		
		if (g.getParentMoveIndex() != -1) {
			list.add(new SolutionStep(g.getParentMoveIndex(), g.getParentMovePermutation()));
		}
	}
	
}