package com.analogbit.puzzletools.client.slidingblocks.state;

import java.util.ArrayList;
import java.util.Arrays;

import com.analogbit.puzzletools.client.util.Color;
import com.analogbit.puzzletools.client.util.DeepCloneable;

public class Grid implements DeepCloneable, Comparable<Grid> {
	
	public static final int MAX_WIDTH = 32;
	public static final int MAX_HEIGHT = 32;
	
	/**
	 * Extremely volatile cache, shared amongst all grids. 
	 */
	private static final int [] gridCache = new int [MAX_WIDTH*MAX_HEIGHT];
	private static Grid gridCached;
	
	private ArrayList<Block> blocks = new ArrayList<Block>();
	
	private int width;
	private int height;
	private int depth;
	private int [] blockAliases;
	
	private Grid parent;
	private int parentMoveIndex;
	private int parentMovePermutation;
	private int distFromGoal;
	
	public Grid(int width, int height) {
		this(width, height, null);
	}
	
	public Grid(int width, int height, ArrayList<Block> blocks) {
		this(width, height, blocks, 0, null);
	}
	
	public Grid(int width, int height, ArrayList<Block> blocks, int depth, Grid previous) {
		if (blocks == null) {
			blocks = new ArrayList<Block>();
		}
		
		this.blocks = blocks;
		this.parent = previous;
		this.width = width;
		this.height = height;
		this.depth = depth;
		this.parentMoveIndex = -1;
		this.parentMovePermutation = -1;
	}
	
	public void setDistanceFromGoal(int distance) {
		this.distFromGoal = distance;
	}
	
	public int getDistanceFromGoal() {
		return distFromGoal;
	}
	
	public void recomputeBlockAliases(GoalState goalState) {
		blockAliases = new int[blocks.size()];
		//reset all the block aliases to their index
		for (int i=0; i<blocks.size(); i++) {
			blockAliases[i] = i;
		}
		
		//merge duplicate block types
		for (int i=0; i<blocks.size()-1; i++) {
			Block block1 = blocks.get(i);
			if (!goalState.mayBeAliased(this, block1)) {
				continue;
			}
			BlockShape shape1 = block1.getBlockShape();
			for (int j=i+1; j<blocks.size(); j++) {
				Block block2 = blocks.get(j);
				if (!goalState.mayBeAliased(this, block2)) {
					continue;
				}
				BlockShape shape2 = block2.getBlockShape();
				
				if (shape1.equals(shape2)) {
					blockAliases[j] = blockAliases[i];
				}
			}
		}
	}
	
	/**
	 * @param dest list to add new grids
	 */
	public void getAllValidPermutations(ArrayList<Grid> dest) {
		for (int i = 0; i < blocks.size(); i++) {
			Block b2m = blocks.get(i);
			
			if (b2m.isImmobile()) {
				continue;
			}

			for (int p = 0; p < 4; p++) {
				Block shifted = b2m.cloneShifted(SolutionStep.SHIFT_PERMUTE_X[p], SolutionStep.SHIFT_PERMUTE_Y[p]);
				if (blockFits(shifted) && !blocksIntersect(blocks, shifted, i)) {
					ArrayList<Block> newArr = new ArrayList<Block>(blocks.size());
					for (int j = 0; j < blocks.size(); j++) {
						if (i == j) {
							newArr.add(shifted);
						} else {
							newArr.add((Block) blocks.get(j).deepClone());
						}
					}
					Grid newGrid = new Grid(width, height, newArr, depth+1, this);
					newGrid.blockAliases = this.blockAliases;
					newGrid.setParentMove(i, p);
					dest.add(newGrid);
				}
			}
		}
	}
	
	private void setParentMove(int index, int permutation) {
		parentMoveIndex = index;
		parentMovePermutation = permutation;
	}
	
	public Grid getParentGrid() {
		return parent;
	}
	
	public int getParentMoveIndex() {
		return parentMoveIndex;
	}
	
	public int getParentMovePermutation() {
		return parentMovePermutation;
	}

	private boolean blocksIntersect(ArrayList<Block> oldArr, Block newBlock, int excludeIndex) {
		for (int i=0; i<oldArr.size(); i++) {
			if (excludeIndex == i) {
				continue;
			}
			if (Block.intersects(oldArr.get(i), newBlock)) {
				return true;
			}
		}
		return false;
	}

	private boolean blockFits(Block block) {
		return block.getLeft() >= 0 && block.getTop() >= 0 && block.getRight() <= getWidth() && block.getBottom() <= getHeight();
	}

	/**
	 * Fills in the cache
	 * @return false if there was block overlap
	 */
	public boolean fillCache() {
		if (gridCached == this) {
			return true;
		}
		gridCached = null;
		Arrays.fill(gridCache, -1);
		for (int i = 0; i < blocks.size(); i++) {
			Block block = blocks.get(i);
			boolean [] blockBits = block.getBlockBits();
			for (int y = 0; y < block.getHeight(); y++) {
			int bitIndexY = y * block.getWidth();
			int gridIndexY = (block.getTop() + y) * width;
			for (int x = 0; x < block.getWidth(); x++) {
				int bitIndex = bitIndexY + x;
				if (blockBits[bitIndex]) {
					int gridIndex = block.getLeft() + x + gridIndexY;
					if (gridCache[gridIndex] != -1) {
						// overlap
						return false;
					}
					
					if (blockAliases != null) {
						//put the ALIAS in the grid so we can eliminate duplicate states
						gridCache[gridIndex] = blockAliases[i];
					} else {
						gridCache[gridIndex] = i;
					}
				}
			}
			
			/*for (int y = block.getTop(); y < block.getBottom(); y++) {
				int indexy = y * width;
				for (int x = block.getLeft(); x < block.getRight(); x++) {
					int bitIndex = indexy + x;
					if (blockBits[bitIndex]) {
						if (gridCache[bitIndex] != -1) {
							// overlap
							return false;
						}
						
						if (blockAliases != null) {
							//put the ALIAS in the grid so we can eliminate duplicate states
							gridCache[bitIndex] = blockAliases[i];
						} else {
							gridCache[bitIndex] = i;
						}
					}
				}*/
			}
		}
		gridCached = this;
		return true;
	}
	
	@Override
	public boolean equals(Object obj) {
		/*Grid gOther = (Grid)obj;
		
		for (int i=0; i<currentState.length; i++) {
			if (!currentState[i].equalsNoCast(gOther.currentState[i])) {
				return false;
			}
		}
		return true;*/
		
		//for some odd reason string compares are much faster than array accesses 
		//in javascript
		return toString().equals(obj.toString());
	}

	@Override
	public int hashCode() {
		/*
		int hash = 0;
		for (int i=0; i<currentState.length; i++) {
			hash ^= 257*(i+1)*currentState[i].hashCode();
		}
		//System.out.println("hashcode:" +hash);
		return hash;
		//*/
		
		//string hash is very strong
		//System.out.println("hashcode:" + hashString().hashCode());
		return hashString().hashCode();
	}
	
	public String toString() {
		if (!fillCache()) {
			return "Bad State";
		}
		String out ="";
		for (int y=0; y<height; y++) {
			for (int x=0; x<width; x++) {
				if (gridCache[y*width + x] == -1) {
					out += "   ";
				} else {
					out += "[" + gridCache[y*width + x] + "]";
				}
			}
			out += "\n";
		}
		return out;
	}
	
	public final String hashString() {
		String out = "";
		if (!fillCache()) {
			return out;
		}
		
		int numCells = width*height;
		for (int i=0; i<numCells; i++) {
			out += gridCache[i];
		}
		return out;
	}
	/* its really difficult to beat string hash by just making stuff up 
	public final int cacheHash() {
		if (!fillCache()) {
			return 0;
		}
		int numCells = width*height;
		int hash = numCells*31;
		for (int i=0; i<numCells; i++) {
			hash = (hash + gridCache[i]) ^ ((i+gridCache[i])*61);
		}
		return hash;
	} */

	public void printSolution() {
		System.out.println("Solution follows: \n=======");
		Grid toPrint = this;
		do {
			System.out.println(toPrint);
			System.out.println();
			toPrint = toPrint.parent;
		} while (toPrint != null);
	}

	public int getHeight() {
		return height;
	}
	
	public int getWidth() {
		return width;
	}

	protected int getDepth() {
		return depth;
	}
	
	protected int getCost() {
		return getDepth() + getDistanceFromGoal();
	}

	public ArrayList<Block> getBlocks() {
		return blocks;
	}

	public Block getBlockAt(int x, int y) {
		for (int i=0; i<blocks.size(); i++) {
			if (blocks.get(i).hasCellAt(x, y)) {
				return blocks.get(i);
			}
		}
		return null;
	}

	/*public int getBlockIndexAt(int x, int y) {
		for (int i=0; i<blocks.size(); i++) {
			if (blocks.get(i).hasCellAt(x, y)) {
				return i;
			}
		}
		return -1;
	}*/
	
	public void addBlock(Block block) {
		blockAliases = null;
		blocks.add(block);
		invalidateCache();
	}

	private void invalidateCache() {
		if (gridCached == this) {
			gridCached = null;
		}
	}
	
	public boolean removeBlock(Block blockRemoving) {
		//go from the end towards the front for speed, also avoids .equals()
		//which cause unwanted behavior
		boolean removed = false;
		for (int i=blocks.size()-1; i>=0; i--) {
			if (blocks.get(i) == blockRemoving) {
				blocks.remove(i);
				removed = true;
				break;
			}
		}

		if (removed) {
			invalidateCache();
			blockAliases = null;
		}
		return removed;
	}

	public boolean addBlockIfLegal(Block newBlock) {
		assert(fillCache());
		
		if (newBlock.getLeft() < 0 || newBlock.getRight() > getWidth() || 
				newBlock.getTop() < 0 || newBlock.getBottom() > getHeight()) {
			return false;
		}
		
		//temporarily add the block 
		addBlock(newBlock);
		//check for overlap
		boolean result = fillCache();
		if (!result) { //there was overlap, undo the add
			removeBlock(newBlock);
		}
		return result;
	}
	
	/**
	 * Replaces a block with a new one.
	 * @param oldBlock The block to replace
	 * @param newBlock The block to replace it with
	 * @return true if a block is replaced
	 */
	public boolean replaceBlockIfLegal(Block oldBlock, Block newBlock) {
		assert(fillCache());
		if (!removeBlock(oldBlock)) {
			return false;
		}
		if (addBlockIfLegal(newBlock)) {
			return true;
		} else {//the add failed, add the unmodified block back in
			addBlock(oldBlock);
			return false;
		}
	}

	public boolean hasMovableBlocks() {
		for (int i=0; i<blocks.size(); i++) {
			if (!blocks.get(i).isImmobile()) {
				return true;
			}
		}
		return false;
	}

	public boolean moveBlockIfLegal(Block block, int x, int y) {
		assert(fillCache());
		if (x < 0 || x + block.getWidth() > getWidth()) {
			return false;
		}
		if (y < 0 || y + block.getHeight() > getHeight()) {
			return false;
		}

		int oldX = block.getX();
		int oldY = block.getY();

		
		block.setX(x);
		block.setY(y);
		
		invalidateCache();
		if (fillCache()) {
			return true;
		} else {
			block.setX(oldX);
			block.setY(oldY);
			return false;
		}
	}

	public void setBlockColor(Block block, Color color) {
		block.setColor(color);
	}

	public void setBlockImmobile(Block block, boolean immobile) {
		block.setImmobile(immobile);
	}

	public GridWithListeners createShallowImmobilityGrid(boolean mobile) {
		ArrayList<Block> keyBlocks = new ArrayList<Block>();
		for (Block b : blocks) {
			if (b.isImmobile() == mobile) {
				keyBlocks.add((Block)b);
			}
		}
		return new GridWithListeners(this.width, this.height, keyBlocks);
	}

	public Grid deepClone() {
		Grid g = new Grid(width, height);
		for (Block b:blocks) {
			g.addBlock((Block) b.deepClone());
		}
		return g;
	}
	
	public void stepForward(SolutionStep step) {
		int moveX = SolutionStep.SHIFT_PERMUTE_X[step.getPermutation()];
		int moveY = SolutionStep.SHIFT_PERMUTE_Y[step.getPermutation()];
		
		Block block = blocks.get(step.getIndexMoved());
		
		moveBlockIfLegal(block, block.getX() + moveX,  block.getY() + moveY);
	}

	public void stepBackward(SolutionStep step) {
		int moveX = -SolutionStep.SHIFT_PERMUTE_X[step.getPermutation()];
		int moveY = -SolutionStep.SHIFT_PERMUTE_Y[step.getPermutation()];
		
		Block block = blocks.get(step.getIndexMoved());
		moveBlockIfLegal(block, block.getX() + moveX,  block.getY() + moveY);
	}

	public int compareTo(Grid o) {
		return this.getCost() - o.getCost();
	}

}