package com.analogbit.puzzletools.client.util;

import com.google.gwt.user.client.Random;

/**
 * RGB color with some HSV support. 
 */
public class Color implements DeepCloneable {
	private int red;
	private int green;
	private int blue;
	
	public static Color random() {
		return new Color(Random.nextInt()&0xFF, Random.nextInt()&0xFF, Random.nextInt()&0xFF);
	}
	
	public static Color randomBright() {
		double hue = Random.nextDouble();
		double sat = Random.nextDouble()/2 + 0.5;
		double val = Random.nextDouble()/4 + 0.75;
		return fromHueSatValue(hue, sat, val);
	}
	
	public Color() {
		this(0, 0, 0);
	}
	
	public Color(Color copy) {
		this(copy.red, copy.green, copy.blue);
	}
	
	public Color(int red, int green, int blue) {
		setColor(red, green, blue);
	}
	
	public Color(double redPercent, double greenPercent, double bluePercent) {
		setColor(redPercent, greenPercent, bluePercent);
	}
	
	//kind of ripped from http://www.easyrgb.com/
	public static Color fromHueSatValue(double hue, double sat, double value) {
		if (sat <= 0) {
			int val = (int) Math.round(value*255);
			return new Color(val, val, val);
		} else {
			double var_h, var_i, var_1, var_2, var_3, r, g, b;
			var_h = hue * 6;
			if ( var_h == 6 ) var_h = 0;      //H must be < 1
			var_i = (int)var_h;               //Or ... var_i = floor( var_h )
			var_1 = value * ( 1 - sat );
			var_2 = value * ( 1 - sat * ( var_h - var_i ) );
			var_3 = value * ( 1 - sat * ( 1 - ( var_h - var_i ) ) );
	
			if      ( var_i == 0 ) { r = value ; g = var_3 ; b = var_1; }
			else if ( var_i == 1 ) { r = var_2 ; g = value ; b = var_1; }
			else if ( var_i == 2 ) { r = var_1 ; g = value ; b = var_3; }
			else if ( var_i == 3 ) { r = var_1 ; g = var_2 ; b = value; }
			else if ( var_i == 4 ) { r = var_3 ; g = var_1 ; b = value; }
			else                   { r = value ; g = var_1 ; b = var_2; }
		
			return new Color(r, g, b);
		}
	}
	
	public void setColor(int red, int green, int blue) {
		this.red   = Math.min(0xFF, red);
		this.green = Math.min(0xFF, green);
		this.blue  = Math.min(0xFF, blue);
		
		this.red   = Math.max(0, this.red);
		this.green = Math.max(0, this.green);
		this.blue  = Math.max(0, this.blue);
	}
	
	public void setColor(double redPercent, double greenPercent, double bluePercent) {
		setColor((int)(redPercent*255), (int)(greenPercent*255), (int)(bluePercent*255));
	}
	
	private static final char nibbleToChar(int nibble) {
		final char lookup[] = {
			'0', '1', '2', '3',
			'4', '5', '6', '7',
			'8', '9', 'A', 'B',
			'C', 'D', 'E', 'F'
		};
		return lookup[nibble];
	}
	
	public void modulate(double amount) {
		setColor((int)(red*amount), (int)(green*amount), (int)(blue*amount));
	}
	
	public void brighten(double amount) {
		int intAmt = (int)(amount*255);
		add(intAmt, intAmt, intAmt);
	}
	
	public void add(int red, int green, int blue) {
		setColor(this.red + red, this.green + green, this.blue + blue);
	}
	
	public String toString() {
		return "#" + nibbleToChar(red >> 4) + nibbleToChar(red & 0x7) +
			nibbleToChar(green >> 4) + nibbleToChar(green & 0x7) +
			nibbleToChar(blue >> 4) + nibbleToChar(blue & 0x7);
	}

	public Object deepClone() {
		return new Color(this);
	}

	public void getHueSatValue(double hsv[]) {
		double R = ( red / 255.0 );
		double G = ( green / 255.0 );
		double B = ( blue / 255.0 );

		double min = Math.min(R, Math.min(G, B));
		double max = Math.max(R, Math.max(G, B));
		double delta = max - min;

		double hue;
		double sat;
		double value = max;
		
		//handle grays
		if (delta == 0) {
		   hue = 0;
		   sat = 0;
		} else {
		   sat = delta / max;

		   double deltaR = (((max - R)/6) + delta*0.5) / delta;
		   double deltaG = (((max - G)/6) + delta*0.5) / delta;
		   double deltaB = (((max - B)/6) + delta*0.5) / delta;

		   if (R == max) {
			   hue = deltaB - deltaG;
		   } else if (G == max) {
			   hue = (1.0 / 3.0) + deltaR - deltaB;
		   } else { //B == max
			   hue = (2.0 / 3.0) + deltaG - deltaR;
		   }

		   if (hue < 0) {
			   hue += 1;
		   } else if (hue > 1) {
			   hue -= 1;
		   }
		}

		hsv[0] = hue;
		hsv[1] = sat;
		hsv[2] = value;
	}
	
}



